// Node editor UI element

#[allow(dead_code)]

use glium::glutin::window::Window as GlutinWindow;
use glium::glutin::dpi::LogicalPosition;
use imgui::{Ui, MouseButton};
use imgui::color::ImColor32;
use imgui::draw_list::DrawListMut;

pub struct DragArea {
    pub offset: [f32; 2],
    pub size: [f32; 2],
    pub border_color: ImColor32,
    pub default_offset: [f32; 2],
    default_size: [f32; 2],
    active: bool
}

pub struct DragAreaMut<'a> {
    pub draw_list: &'a DrawListMut<'a>,
    pub area_min: [f32; 2],
    pub area_max: [f32; 2],
    pub offset: &'a mut [f32; 2],
    pub hovered: bool
}

impl DragArea {
    pub fn new(size: [f32; 2], offset: [f32; 2]) -> DragArea {
        DragArea {
            offset,
            size,
            border_color: ImColor32::from_rgb(15u8, 20u8, 25u8),
            default_offset: offset,
            default_size: size,
            active: false
        }
    }

    pub fn build
        <F: FnMut(DragAreaMut)>
        (&mut self, ui: &Ui, window: &GlutinWindow, mut build_children: F)
    {
        let mouse_delta = ui.io().mouse_delta;
        let mouse_pos = ui.io().mouse_pos;
        
        if ui.button("reset view") {
            self.offset = self.default_offset;
        }

        ui.invisible_button("##empty", self.size);

        let rect_min = ui.item_rect_min();
        let rect_max = ui.item_rect_max();

        let mmb_down = ui.is_mouse_dragging(MouseButton::Middle);
        let hovered = ui.is_item_hovered();

        if !self.active && mmb_down && hovered {
            self.active = true;
        } else if self.active && !mmb_down {
            self.active = false;
        }

        if self.active {
            if rect_min[0] > mouse_pos[0] {
                window.set_cursor_position(
                    LogicalPosition { x: rect_max[0], y: mouse_pos[1] }
                ).unwrap();
                self.offset[0] -= self.size[0];
            }

            if rect_max[0] < mouse_pos[0] {
                window.set_cursor_position(
                    LogicalPosition { x: rect_min[0], y: mouse_pos[1] }
                ).unwrap();
                self.offset[0] += self.size[0];
            }

            if rect_min[1] > mouse_pos[1] {
                window.set_cursor_position(
                    LogicalPosition { x: mouse_pos[0], y: rect_max[1] }
                ).unwrap();
                self.offset[1] -= self.size[1];
            }

            if rect_max[1] < mouse_pos[1] {
                window.set_cursor_position(
                    LogicalPosition { x: mouse_pos[0], y: rect_min[1] }
                ).unwrap();
                self.offset[1] += self.size[1];
            }
            
            self.offset[0] += mouse_delta[0];
            self.offset[1] += mouse_delta[1];
        }

        let draw_list = ui.get_window_draw_list();

        draw_list.with_clip_rect(rect_min, rect_max, || {
            draw_list.add_rect(rect_min, rect_max, self.border_color).build();
            build_children(
                DragAreaMut {
                    draw_list: &draw_list,
                    area_min: rect_min,
                    area_max: rect_max,
                    offset: &mut self.offset,
                    hovered
                }
            );
        });

        // ui.set_cursor_screen_pos([rect_min[0], rect_max[1] + 10.0]);
    }
}
